extern "C" {
#include <display_syscalls.h>
#include <keyboard_syscalls.h>
#include <keyboard.hpp>
#include <color.h>
#include <misc_tools.h>
#include <HEAP_syscalls.h>
#include <RTC_syscalls.h>
#include <CONVERT_syscalls.h>
//
#define TEXT_SCREEN_WIDTH	21
#define TEXT_SCREEN_HEIGHT	8

#define MIN(a, b)	a < b ? a : b
#define MAX(a, b)	a > b ? a : b

static const int FILLED		= -1;
static const int FLAGGED	= -2;

int numMines = 20;
int fieldWidth = TEXT_SCREEN_WIDTH;
int fieldHeight = TEXT_SCREEN_HEIGHT;

int maxMines = 99;
int minMines = 5;
int maxWidth = TEXT_SCREEN_WIDTH;
int minWidth = 5;
int maxHeight = TEXT_SCREEN_HEIGHT;
int minHeight = 5;
	
char visibleField[TEXT_SCREEN_WIDTH][TEXT_SCREEN_HEIGHT];
char mineField[TEXT_SCREEN_WIDTH][TEXT_SCREEN_HEIGHT];
char* filled	= "  \xE6\xA4";
char* flagged	= "  \xE6\xD3";
char* mine		= "  \xE6\xA0";
char* star		= "  \xE5\xAD";

char* zero		= "  \xE6\xA3";
char* one		= "  \xE5\xF0";
char* two		= "  \xE5\xF1";
char* three		= "  \xE5\xF2";
char* four		= "  \xE5\xF3";
char* five		= "  \xE5\xF4";
char* six		= "  \xE5\xF5";
char* seven		= "  \xE5\xFB";
char* eight		= "  \xE5\xFC";

char* rightArrow	= "  \xE5\xE9";

const unsigned int default_fkeys[] = 
  { 0x0000FFFF,0,0x0000FFFF,0,0x0000FFFF,0,0x0000FFFF,0,0x0000FFFF,0,0x0000FFFF,0 };

void DrawField();
void CreateFields();
char* getVisibleType(int i, int j);
int Cascade(int x, int y);
void Die();
void Win();
void PrintXYVisibleSpot(int i, int j);
void About();
bool Settings();
char* concat(char* str1, char* str2);

void G3A_main(void) {  
	Set_FKeys1( 0, (unsigned int*)default_fkeys );
	Set_FKeys2( 0 );
start:
	int total = 0;
	int x = 1;
	int y = 1;
	int key;
	int ticks = RTC_GetTicks();
	random(ticks);
	CreateFields();

	// Strings that get printed from PrintXY need two spaces of padding...

	while (1) {
		Bdisp_AllCr_VRAM();
		DrawField();
		PrintXY(x, y, getVisibleType(x-1, y-1), TEXT_MODE_INVERT, TEXT_COLOR_BLUE);
		GetKey(&key);
		switch (key) {
			case KEY_CTRL_UP:
				y = y == 1 ? fieldHeight : y-1;
				break;
			case KEY_CTRL_DOWN:
				y = y == fieldHeight ? 1 : y+1;
				break;
			case KEY_CTRL_LEFT:
				x = x == 1 ? fieldWidth : x-1;
				break;
			case KEY_CTRL_RIGHT:
				x = x == fieldWidth ? x=1 : x+1;
				break;
			case KEY_CTRL_F2:
				if (visibleField[x-1][y-1] == FILLED) {
					visibleField[x-1][y-1] = FLAGGED;
				} else if (visibleField[x-1][y-1] == FLAGGED) {
					visibleField[x-1][y-1] = FILLED;
				}
				break;
			case KEY_CTRL_F1:
				if (visibleField[x-1][y-1] == FILLED) {
					int num = Cascade(x-1,y-1);
					if (!num) {
						Die();
						goto start;
					}
					total += num;
					if (total == fieldHeight*fieldWidth - numMines) {
						Win();
						goto start;
					}
				}
				break;
			case KEY_CTRL_F3:
				About();
				break;
			case KEY_CTRL_F5:
				if (Settings()) {
					goto start;
				}
				break;
			case KEY_CTRL_F6:
				goto start;
				break;
		}
	}

	return;
}

void About() {
	Bdisp_AllCr_VRAM();
	PrintXY(1, 1, "  About Minesweeper", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
	PrintXY(1, 2, "  v1.0 by:", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
	PrintXY(1, 3, "  Merthsoft Creations", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
	PrintXY(1, 4, "  (Shaun McFall, 2011)", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
	PrintXY(1, 6, "  Thanks:", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
	PrintXY(1, 7, "  Fishbot - Prizm Stuff", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
	PrintXY(1, 8, "  KermM - C Stuff, Art", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);

	int key;
	do {
		GetKey(&key);
	} while (!key);
}

bool Settings() {
	int key;
	int y = 1;
	unsigned char out[3];

	int newValues[] = {numMines, fieldWidth, fieldHeight};
	int maxValues[] = {maxMines, maxWidth, maxHeight};
	int minValues[] = {minMines, minWidth, minHeight};

	while (1) {
		Bdisp_AllCr_VRAM();

		itoa(newValues[0], out);
		PrintXY(2, 1,  concat("  Number of Mines: ", (char*)out), TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
		itoa(newValues[1], out);
		PrintXY(2, 2,  concat("  Field Width:     ", (char*)out), TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
		itoa(newValues[2], out);
		PrintXY(2, 3,  concat("  Field Height:    ", (char*)out), TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);

		PrintXY(1, 8, "  BACK              NEW", TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);

		PrintXY(1, y, rightArrow, TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);

		GetKey(&key);
		switch (key) {
			case KEY_CTRL_UP:
				y = y == 1 ? 3 : y-1;
				break;
			case KEY_CTRL_DOWN:
				y = y == 3 ? 1 : y+1;
				break;
			case KEY_CTRL_LEFT:
				newValues[y-1] = MAX(newValues[y-1]-1, minValues[y-1]);
				// Max mines is total size - 1
				maxValues[0] = newValues[1] * newValues[2] - 1;
				newValues[0] = MIN(newValues[0], maxValues[0]);
				break;
			case KEY_CTRL_RIGHT:
				newValues[y-1] = MIN(newValues[y-1]+1, maxValues[y-1]);
				// Max mines is total size - 1
				maxValues[0] = newValues[1] * newValues[2] - 1;
				newValues[0] = MIN(newValues[0], maxValues[0]);
				break;
			case KEY_CTRL_F1:
				return 0;
				break;
			case KEY_CTRL_F6:
				numMines = newValues[0];
				fieldWidth = newValues[1];
				fieldHeight = newValues[2];
				return 1;
				break;
		}
	}
}

char* concat(char* str1, char* str2) {
	int len1 = 0;
	int len2 = 0;
	for (;str1[len1] != '\0'; len1++);
	for (;str2[len2] != '\0'; len2++);

	char* str3 = (char*)malloc(len1 + len2 + 1);
	int i;
	for (i = 0; i < len1; str3[i] = str1[i], i++);
	for (i = 0; i < len2; str3[len1+i] = str2[i], i++);
	str3[len1 + len2] = 0;
	
	return str3;
}

void PrintXYVisibleSpot(int i, int j) {
	PrintXY(i, j, getVisibleType(i-1,j-1), TEXT_MODE_NORMAL, visibleField[i-1][j-1] == FLAGGED?TEXT_COLOR_GREEN:TEXT_COLOR_BLACK);
}

void DrawField() {
	int i, j;
	for (i = 1; i <= fieldWidth; i++) {
		for (j = 1; j <= fieldHeight; j++) {
			PrintXYVisibleSpot(i, j);
		}
	}
}

void Win() {
	int i, j;
	for (i = 1; i <= fieldWidth; i++) {
		for (j = 1; j <= fieldHeight; j++) {
			PrintXYVisibleSpot(i, j);
			if (mineField[i-1][j-1]) {
				PrintXY(i, j, star, TEXT_MODE_INVERT, TEXT_COLOR_YELLOW);
			}
		}
	}
	
	int key;
	do {
		GetKey(&key);
	} while (key != KEY_CTRL_F6);
}

void Die() {
	int i, j;
	for (i = 1; i <= fieldWidth; i++) {
		for (j = 1; j <= fieldHeight; j++) {
			PrintXY(i, j, getVisibleType(i-1,j-1), TEXT_MODE_NORMAL, TEXT_COLOR_BLACK);
			if (mineField[i-1][j-1]) {
				PrintXY(i, j, mine, visibleField[i-1][j-1] == FLAGGED?TEXT_MODE_INVERT:TEXT_MODE_NORMAL, TEXT_COLOR_RED);
			}
		}
	}
	
	int key;
	do {
		GetKey(&key);
	} while (key != KEY_CTRL_F6);
}

int Cascade(int x, int y) {
	if (x < 0 || y < 0 || x >= fieldWidth || y >= fieldHeight)
      return 0;
    if (visibleField[x][y] >= 0 || visibleField[x][y] == -2)
      return 0;
    if (mineField[x][y] == 1)
      return 0;
    int neighbors = 0;
	int i, j;
    for (i = -1; i <= 1; i++) {
      for (j = -1; j <= 1; j++) {
        if (!(x+i < 0 || y+j < 0 || x+i >= fieldWidth || y+j >= fieldHeight))
          neighbors += mineField[x+i][y+j];
      }
    }
    visibleField[x][y] = neighbors;
    int num = 1;
    if (neighbors == 0) {
      for (i = -1; i <= 1; i++) {
        for (j = -1; j <= 1; j++) {
          num += Cascade(x + i, y + j);
        }
      }
    }
    return num;
}

void CreateFields() {
	int i, j;
	
	for (i = 0; i < fieldWidth; i++) {
		for (j = 0; j < fieldHeight; j++) {
			visibleField[i][j] = FILLED;
			mineField[i][j] = 0;
		}
	}

	for (i = 0; i < numMines; i++) {
		int x, y;
		do {
			x = random() % fieldWidth;
			y = random() % fieldHeight;
		} while (mineField[x][y] != 0);
		mineField[x][y] = 1;
	}
}

char* getVisibleType(int i, int j) {
	char* type;
	switch (visibleField[i][j]) {
		case FILLED:
			type = filled;
			break;
		case FLAGGED:
			type = flagged;
			break;
		case 0:
			type = zero;
			break;
		case 1:
			type = one;
			break;
		case 2:
			type = two;
			break;
		case 3:
			type = three;
			break;
		case 4:
			type = four;
			break;
		case 5:
			type = five;
			break;
		case 6:
			type = six;
			break;
		case 7:
			type = seven;
			break;
		case 8:
			type = eight;
			break;
	}
	return type;
}

#pragma section _BR_Size
unsigned long BR_Size;
#pragma section

}

